'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _slicedToArray2 = require('babel-runtime/helpers/slicedToArray');

var _slicedToArray3 = _interopRequireDefault(_slicedToArray2);

exports.observer = observer;

var _react = require('react');

var _observerUtil = require('@nx-js/observer-util');

var _subscribe = require('../subscribe');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function observer(baseComponent) {
  var baseComponentName = baseComponent.displayName || baseComponent.name;
  // memo; we are not intested in deep updates
  // in props; we assume that if deep objects are changed,
  // this is in observables, which would have been tracked anyway

  var memoComponent = (0, _react.memo)(function (props) {
    // forceUpdate 2.0
    var forceUpdate = useForceUpdate();

    // wrap the baseComponent into an observe decorator once.
    // This way it will track any observable changes and will trigger rerender
    var observedComponent = (0, _react.useMemo)(function () {
      var update = function update() {
        // TODO: Decide whether the check for unmount is needed here
        forceUpdate();
      };
      var batchedUpdate = function batchedUpdate() {
        return _subscribe.batching.add(update);
      };
      return (0, _observerUtil.observe)(baseComponent, {
        scheduler: batchedUpdate,
        lazy: true
      });
    }, []);

    // clean up observer on unmount
    useUnmount(function () {
      return (0, _observerUtil.unobserve)(observedComponent);
    });

    return observedComponent(props);
  });
  memoComponent.displayName = baseComponentName;
  if (baseComponent.propTypes) {
    memoComponent.propTypes = baseComponent.propTypes;
  }
  return memoComponent;
} // ref: https://github.com/mobxjs/mobx-react-lite/blob/master/src/observer.ts


function useForceUpdate() {
  var _useState = (0, _react.useState)(1),
      _useState2 = (0, _slicedToArray3.default)(_useState, 2),
      tick = _useState2[0],
      setTick = _useState2[1];

  return function () {
    setTick(tick + 1);
  };
}

// TODO: Might change to just `useEffect` in future. Don't know which one fits here better yet.
function useUnmount(fn) {
  (0, _react.useLayoutEffect)(function () {
    return fn;
  }, []);
}